#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import contextlib

from mock import create_autospec, patch
from pystachio.config import Config
from twitter.common.contextutil import temporary_file

from apache.aurora.client.commands.core import create
from apache.aurora.config import AuroraConfig

from .util import AuroraClientCommandTest

from gen.apache.aurora.api.ttypes import (
    AssignedTask,
    JobKey,
    ScheduledTask,
    ScheduleStatus,
    ScheduleStatusResult,
    TaskEvent,
    TaskQuery
)


class TestClientCreateCommand(AuroraClientCommandTest):
  @classmethod
  def setup_mock_options(cls):
    """set up to get a mock options object."""
    mock_options = create_autospec(spec={}, instance=True)
    mock_options.json = False
    mock_options.bindings = {}
    mock_options.open_browser = False
    mock_options.cluster = None
    mock_options.wait_until = 'RUNNING'  # or 'FINISHED' for other tests
    mock_options.disable_all_hooks = False
    mock_options.disable_all_hooks_reason = None
    return mock_options

  @classmethod
  def create_mock_task(cls, task_id, instance_id, initial_time, status):
    mock_task = create_autospec(spec=ScheduledTask, instance=True)
    mock_task.assignedTask = create_autospec(spec=AssignedTask, instance=True)
    mock_task.assignedTask.taskId = task_id
    mock_task.assignedTask.instanceId = instance_id
    mock_task.status = status
    mock_task_event = create_autospec(spec=TaskEvent, instance=True)
    mock_task_event.timestamp = initial_time
    mock_task.taskEvents = [mock_task_event]
    return mock_task

  @classmethod
  def create_mock_status_query_result(cls, scheduleStatus):
    mock_query_result = cls.create_simple_success_response()
    mock_query_result.result.scheduleStatusResult = create_autospec(
        spec=ScheduleStatusResult,
        spec_set=False,
        instance=True)
    mock_task_one = cls.create_mock_task('hello', 0, 1000, scheduleStatus)
    mock_task_two = cls.create_mock_task('hello', 1, 1004, scheduleStatus)
    mock_query_result.result.scheduleStatusResult.tasks = [mock_task_one, mock_task_two]
    return mock_query_result

  @classmethod
  def create_mock_query(cls):
    return TaskQuery(
        jobKeys=[JobKey(role=cls.TEST_ROLE, environment=cls.TEST_ENV, name=cls.TEST_JOB)])

  @classmethod
  def get_createjob_response(cls):
    # Then, we call api.create_job(config)
    return cls.create_simple_success_response()

  @classmethod
  def get_failed_createjob_response(cls):
    return cls.create_error_response()

  @classmethod
  def assert_create_job_called(cls, mock_api):
    # Check that create_job was called exactly once, with an AuroraConfig parameter.
    assert mock_api.create_job.call_count == 1
    assert isinstance(mock_api.create_job.call_args_list[0][0][0], AuroraConfig)

  @classmethod
  def assert_scheduler_called(cls, mock_api, mock_query, num_queries):
    # scheduler.scheduler() is called once, as a part of the handle_open call.
    assert mock_api.scheduler_proxy.getTasksWithoutConfigs.call_count == num_queries
    mock_api.scheduler_proxy.getTasksWithoutConfigs.assert_called_with(mock_query)

  def test_simple_successful_create_job(self):
    """Run a test of the "create" command against a mocked-out API:
    Verifies that the creation command sends the right API RPCs, and performs the correct
    tests on the result."""
    mock_options = self.setup_mock_options()

    # create first calls get_job_config, which calls get_config. As long as we've got the options
    # set up correctly, this should work.

    # Next, create gets an API object via make_client. We need to replace that with a mock API.
    (mock_api, mock_scheduler_proxy) = self.create_mock_api()
    with contextlib.nested(
        patch('apache.aurora.client.commands.core.make_client', return_value=mock_api),
        patch('twitter.common.app.get_options', return_value=mock_options)) as (make_client,
        options):

      # After making the client, create sets up a job monitor.
      # The monitor uses TaskQuery to get the tasks. It's called at least twice:once before
      # the job is created, and once after. So we need to set up mocks for the query results.
      mock_query = self.create_mock_query()
      mock_scheduler_proxy.getTasksWithoutConfigs.side_effect = [
        self.create_mock_status_query_result(ScheduleStatus.RUNNING)
      ]

      # With the monitor set up, create finally gets around to calling create_job.
      mock_api.create_job.return_value = self.get_createjob_response()

      # Finally, it calls the monitor to watch and make sure the jobs started;
      # but we already set that up in the side-effects list for the query mock.

      # This is the real test: invoke create as if it had been called by the command line.
      with temporary_file() as fp:
        fp.write(self.get_valid_config())
        fp.flush()
        create(['west/mchucarroll/test/hello', fp.name])

      # Now check that the right API calls got made.
      # Check that create_job was called exactly once, with an AuroraConfig parameter.
      self.assert_create_job_called(mock_api)
      self.assert_scheduler_called(mock_api, mock_query, 1)
      # make_client should have been called once.
      make_client.assert_called_with('west')

  def test_create_job_wait_until_finished(self):
    """Run a test of the "create" command against a mocked-out API:
    this time, make the monitor check status several times before successful completion.
    """
    mock_options = self.setup_mock_options()
    (mock_api, mock_scheduler_proxy) = self.create_mock_api()
    with contextlib.nested(
        patch('threading._Event.wait'),
        patch('apache.aurora.client.commands.core.make_client', return_value=mock_api),
        patch('twitter.common.app.get_options', return_value=mock_options)) as (sleep, make_client,
        options):
      mock_query = self.create_mock_query()
      mock_options.wait_until = 'FINISHED'
      mock_query_results = [
          self.create_mock_status_query_result(ScheduleStatus.PENDING),
          self.create_mock_status_query_result(ScheduleStatus.PENDING),
          self.create_mock_status_query_result(ScheduleStatus.RUNNING),
          self.create_mock_status_query_result(ScheduleStatus.FINISHED),
      ]
      mock_scheduler_proxy.getTasksWithoutConfigs.side_effect = mock_query_results
      mock_api.create_job.return_value = self.get_createjob_response()
      with temporary_file() as fp:
        fp.write(self.get_valid_config())
        fp.flush()
        create(['west/mchucarroll/test/hello', fp.name])

      # Now check that the right API calls got made.
      # Check that create_job was called exactly once, with an AuroraConfig parameter.
      self.assert_create_job_called(mock_api)
      self.assert_scheduler_called(mock_api, mock_query, 4)
      # make_client should have been called once.
      make_client.assert_called_with('west')

  def test_create_job_failed(self):
    """Run a test of the "create" command against a mocked-out API:
    this time, make the monitor check status several times before successful completion.
    """
    mock_options = self.setup_mock_options()
    (mock_api, mock_scheduler_proxy) = self.create_mock_api()
    with contextlib.nested(
        patch('apache.aurora.client.commands.core.make_client', return_value=mock_api),
        patch('twitter.common.app.get_options', return_value=mock_options)) as (make_client,
        options):
      mock_api.create_job.return_value = self.get_failed_createjob_response()
      # This is the real test: invoke create as if it had been called by the command line.
      with temporary_file() as fp:
        fp.write(self.get_valid_config())
        fp.flush()
        self.assertRaises(SystemExit, create, (['west/mchucarroll/test/hello', fp.name]))

      # Now check that the right API calls got made.
      # Check that create_job was called exactly once, with an AuroraConfig parameter.
      self.assert_create_job_called(mock_api)

      # make_client should have been called once.
      make_client.assert_called_with('west')

  def test_delayed_job(self):
    """Run a test of the "create" command against a mocked-out API:
    this time, make the monitor check status several times before successful completion.
    """
    mock_options = self.setup_mock_options()
    (mock_api, mock_scheduler_proxy) = self.create_mock_api()
    with contextlib.nested(
        patch('threading._Event.wait'),
        patch('apache.aurora.client.commands.core.make_client', return_value=mock_api),
        patch('twitter.common.app.get_options', return_value=mock_options)) as (sleep, make_client,
        options):
      mock_query = self.create_mock_query()
      mock_query_results = [
          self.create_mock_status_query_result(ScheduleStatus.PENDING),
          self.create_mock_status_query_result(ScheduleStatus.PENDING),
          self.create_mock_status_query_result(ScheduleStatus.RUNNING)
      ]
      mock_scheduler_proxy.getTasksWithoutConfigs.side_effect = mock_query_results
      mock_api.create_job.return_value = self.get_createjob_response()
      with temporary_file() as fp:
        fp.write(self.get_valid_config())
        fp.flush()
        create(['west/mchucarroll/test/hello', fp.name])

      # Now check that the right API calls got made.
      self.assert_create_job_called(mock_api)
      self.assert_scheduler_called(mock_api, mock_query, 3)
      # make_client should have been called once.
      make_client.assert_called_with('west')

  def test_create_job_failed_invalid_config(self):
    """Run a test of the "create" command against a mocked-out API, with a configuration
    containing a syntax error"""
    mock_options = self.setup_mock_options()
    (mock_api, mock_scheduler_proxy) = self.create_mock_api()
    with contextlib.nested(
        patch('apache.aurora.client.commands.core.make_client', return_value=mock_api),
        patch('twitter.common.app.get_options', return_value=mock_options)) as (make_client,
        options):
      with temporary_file() as fp:
        fp.write(self.get_invalid_config('invalid_clause=oops'))
        fp.flush()
        self.assertRaises(Config.InvalidConfigError, create, (['west/mchucarroll/test/hello',
            fp.name]))

      # Now check that the right API calls got made.
      # Check that create_job was not called.
      assert mock_api.create_job.call_count == 0

      assert mock_scheduler_proxy.getTasksWithoutConfigs.call_count == 0
      # make_client should not have been called.
      assert make_client.call_count == 0

  def test_create_job_failed_invalid_config_two(self):
    """Run a test of the "create" command against a mocked-out API:
    this time, make the monitor check status several times before successful completion.
    """
    mock_options = self.setup_mock_options()
    (mock_api, mock_scheduler_proxy) = self.create_mock_api()
    with contextlib.nested(
        patch('apache.aurora.client.commands.core.make_client', return_value=mock_api),
        patch('twitter.common.app.get_options', return_value=mock_options)) as (make_client,
        options):
      with temporary_file() as fp:
        fp.write(self.get_invalid_config('invalid_clause=\'oops\','))
        fp.flush()
        self.assertRaises(AttributeError, create, (['west/mchucarroll/test/hello', fp.name]))

      # Now check that the right API calls got made.
      # Check that create_job was not called.
      assert mock_api.create_job.call_count == 0
      # getTasksWithoutConfigs was called once, before the create_job
      assert mock_scheduler_proxy.getTasksWithoutConfigs.call_count == 0
      # make_client should not have been called.
      assert make_client.call_count == 0
